//
// browser.cs: Mono documentation browser
//
// Author:
//   Miguel de Icaza
//
// (C) 2003 Ximian, Inc.
//
// TODO:
//   Add support for printing.
//   Add back/forward buttons.
//   Add search facility
//   Add Index
//
using Gtk;
using GtkSharp;
using Glade;
using System;
using System.IO;
using System.Reflection;
using System.Collections;

class Browser {
	Glade.XML ui;
	Gtk.Window MainWindow;
	
	static int Main (string [] args)
	{
		string topic = null;
		
		for (int i = 0; i < args.Length; i++){
			switch (args [i]){
			case "--html":
				if (i+1 == args.Length){
					Console.WriteLine ("--html needed argument");
					return 1; 
				}
				
				Node n;
				RootTree help_tree = RootTree.LoadTree ();
				string res = help_tree.RenderUrl (args [i+1], out n);
				if (res != null){
					Console.WriteLine (res);
					return 0;
				} else {
					return 1;
				}
			case "--make-index":
				RootTree.MakeIndex ();
				return 0;
				
			case "--help":
				Console.WriteLine ("Options are:\n"+
						   "browser [--html TOPIC] [--make-index] [TOPIC]");
				return 0;
			default:
				topic = args [i];
				break;
			}
			
		}
		Application.Init ();
		Browser browser = new Browser ();

		if (args.Length > 0)
			browser.LoadUrl (topic);
		Application.Run ();
		return 0;
	}

	[Glade.Widget] TreeView reference_tree;
	[Glade.Widget] Box help_container;
	[Glade.Widget] ScrolledWindow html_container;
	[Glade.Widget] Statusbar statusbar;
	[Glade.Widget] Button back_button;
	[Glade.Widget] Button forward_button;
	[Glade.Widget] Frame matches;
	[Glade.Widget] Box search_box;
	
	public History history;

	HTML html;
	TreeBrowser tree_browser;

	RootTree help_tree;

	// For the status bar.
	uint context_id;
	
	//
	// Search
	//
	IndexReader index_reader;
	BigList index_list;
	MatchModel match_model;
	BigList match_list;
	IndexEntry current_entry = null;
	
	Browser ()
	{
		ui = new Glade.XML (null, "browser.glade", "window1", null);
		ui.Autoconnect (this);

		MainWindow = (Gtk.Window) ui["window1"];
		MainWindow.DeleteEvent += new DeleteEventHandler (delete_event_cb);
                
                Stream icon = GetResourceImage ("monodoc.png");

                if (icon != null)
                        MainWindow.Icon = new Gdk.Pixbuf (icon);

		help_tree = RootTree.LoadTree ();
		tree_browser = new TreeBrowser (help_tree, reference_tree, this);

		html = new HTML ();
		html.Show ();
		html_container.Add (html);
                html.LinkClicked += new LinkClickedHandler (LinkClicked);
		html.OnUrl += new OnUrlHandler (OnUrl);
		html.UrlRequested += new UrlRequestedHandler (UrlRequested);
		context_id = statusbar.GetContextId ("");

		history = new History (back_button, forward_button);

		SetupIndex ();
	}

	Stream GetResourceImage (string name)
	{
		Assembly assembly = System.Reflection.Assembly.GetCallingAssembly ();
		System.IO.Stream s = assembly.GetManifestResourceStream (name);
		
		return s;
	}

	void UrlRequested (object sender, UrlRequestedArgs args)
	{
		Console.WriteLine ("Image requested: " + args.Url);
		Stream s = help_tree.GetImage (args.Url);
		
		if (s == null)
			s = GetResourceImage ("monodoc.png");
		byte [] buffer = new byte [8192];
		int n;
		
		while ((n = s.Read (buffer, 0, 8192)) != 0) {
			args.Handle.Write (buffer, n);
		}
	}
	
	public class LinkPageVisit : PageVisit {
		Browser browser;
		string url;
		
		public LinkPageVisit (Browser browser, string url)
		{
			this.browser = browser;
			this.url = url;
		}

		public override void Go ()
		{
			Node n;
			
			string res = browser.help_tree.RenderUrl (url, out n);
			browser.Render (res, n, url);
		}
	}
	
        void LinkClicked (object o, LinkClickedArgs args)
	{
		LoadUrl (args.Url);
	}

	public void LoadUrl (string url)
	{
		if (url.StartsWith("#"))
		{
			// FIXME: This doesn't deal with whether anchor jumps should go in the history
			html.JumpToAnchor(url.Substring(1));
			return;
		}
		
		Node node;
		
		Console.Error.WriteLine ("Trying: {0}", url);
		string res = help_tree.RenderUrl (url, out node);
		if (res != null){
			Render (res, node, url);
			history.AppendHistory (new LinkPageVisit (this, url));
			return;
		}
		
		Console.Error.WriteLine ("+----------------------------------------------+");
		Console.Error.WriteLine ("| Here we should locate the provider for the   |");
		Console.Error.WriteLine ("| link.  Maybe using this document as a base?  |");
		Console.Error.WriteLine ("| Maybe having a locator interface?   The short|");
		Console.Error.WriteLine ("| urls are not very useful to locate types     |");
		Console.Error.WriteLine ("+----------------------------------------------+");
		Render (url, null, url);
	}

	//
	// Renders the HTML text in `text' which was computed from `url'.
	// The Node matching is `matched_node'.
	//
	// `url' is only used for debugging purposes
	//
	public void Render (string text, Node matched_node, string url)
	{
		Gtk.HTMLStream stream = html.BeginContent ("text/html");

		stream.Write ("<html><body>");
		stream.Write (text);
		stream.Write ("</body></html>");

		if (matched_node == null){
			Console.Error.WriteLine ("[{0}] No matching node", url);
			return;
		} else {
			Console.Error.WriteLine ("[{0}] Matched node", url);
			tree_browser.ShowNode (matched_node);
		}
	}
	
	//
	// Invoked when the mouse is over a link
	//
	string last_url = "";
	void OnUrl (object o, OnUrlArgs args)
	{
		string new_url = args.Url;

		if (new_url == null)
			new_url = "";
		
		if (new_url != last_url){
			statusbar.Pop (context_id);
			statusbar.Push (context_id, new_url);
			last_url = new_url;
		}
	}
	
	void delete_event_cb (object o, DeleteEventArgs args)
	{
		Application.Quit ();
	}

	void SetupIndex ()
	{
		index_reader = IndexReader.Load ("monodoc.index");
		if (index_reader == null)
			return;
		
		index_list = new BigList (index_reader);
		index_list.SetSizeRequest (100, 400);

		index_list.ItemSelected += new ItemSelected (OnItemSelected);
		HBox box = new HBox (false, 0);
		box.PackStart (index_list, true, true, 0);
		Scrollbar scroll = new VScrollbar (index_list.Adjustment);
		box.PackEnd (scroll, false, false, 0);
		
		search_box.PackStart (box, true, true, 0);
		box.ShowAll ();

		//
		// Setup the matches.
		//
		match_model = new MatchModel (this);
		matches.Hide ();
		match_list = new BigList (match_model);
		match_list.Show ();
		match_list.ItemSelected += new ItemSelected (OnMatchSelected);
		matches.Add (match_list);
		index_list.SetSizeRequest (100, 200);
	}

	//
	// This class is used as an implementation of the IListModel
	// for the matches for a given entry.
	// 
	public class MatchModel : IListModel {
		Browser parent;
		
		public MatchModel (Browser parent)
		{
			this.parent = parent;
		}
		
		public int Rows {
			get {
				if (parent.current_entry != null){
					return parent.current_entry.Count;
				} else
					return 0;
			}
		}

		public string GetValue (int row)
		{
			if (parent.current_entry != null){
				return parent.current_entry [row].Caption;
			}
			return "Unknown";
		}
	}

	int selected_index = -1;

	//
	// When an item is selected from the main index list
	//
	void OnItemSelected (int index)
	{
		selected_index = index;

		current_entry = index_reader.GetIndexEntry (index);
		if (current_entry.Count > 1){
			matches.Show ();
			match_list.Refresh ();
		} else {
			matches.Hide ();
			LoadUrl (current_entry [0].Url);
		}
	}

	void OnMatchSelected (int index)
	{
		LoadUrl (current_entry [index].Url);
	}
}

//
// This class implements the tree browser
//
class TreeBrowser {
	Browser browser;

	TreeView tree_view;
	
	TreeStore store;
	RootTree help_tree;
	TreeIter root_iter;

	//
	// This hashtable maps an iter to its node.
	//
	Hashtable iter_to_node;

	//
	// This hashtable maps the node to its iter
	//
	Hashtable node_to_iter;

	//
	// Maps a node to its TreeIter parent
	//
	Hashtable node_parent;

	public TreeBrowser (RootTree help_tree, TreeView reference_tree, Browser browser)
	{
		this.browser = browser;
		tree_view = reference_tree;
		iter_to_node = new Hashtable ();
		node_to_iter = new Hashtable ();
		node_parent = new Hashtable ();

		// Setup the TreeView
		tree_view.AppendColumn ("name_col", new CellRendererText (), "text", 0);

		// Bind events
		tree_view.RowExpanded += new GtkSharp.RowExpandedHandler (RowExpanded);
		tree_view.Selection.Changed += new EventHandler (RowActivated);

		// Setup the model
		this.help_tree = help_tree;
		store = new TreeStore (typeof (string));

		root_iter = store.AppendValues ("Mono Documentation");
		iter_to_node [root_iter] = help_tree;
		node_to_iter [help_tree] = root_iter;
		PopulateNode (help_tree, root_iter);

		reference_tree.Model = store;
	}

	void PopulateNode (Node node, TreeIter parent)
	{
		if (node.Nodes == null)
			return;

		TreeIter iter;
		foreach (Node n in node.Nodes){
			iter = store.AppendValues (parent, n.Caption);
			iter_to_node [iter] = n;
			node_to_iter [n] = iter;
		}
	}

	Hashtable populated = new Hashtable ();
	
	void RowExpanded (object o, GtkSharp.RowExpandedArgs args)
	{
		Node result = iter_to_node [args.Iter] as Node;

		Open (result);
	}

	void Open (Node node)
	{
		if (node == null){
			Console.Error.WriteLine ("Expanding something that I do not know about");
			return;
		}

		if (populated.Contains (node))
			return;
		
		//
		// We need to populate data on a second level
		//
		if (node.Nodes == null)
			return;

		foreach (Node n in node.Nodes){
			PopulateNode (n, (TreeIter) node_to_iter [n]);
		}
		populated [node] = true;
	}
	
	void PopulateTreeFor (Node n)
	{
		if (populated [n] == null){
			if (n.Parent != null) {
				OpenTree (n.Parent);
			}
		} 
		Open (n);
	}

	public void OpenTree (Node n)
	{
		PopulateTreeFor (n);

		TreeIter iter = (TreeIter) node_to_iter [n];
		TreePath path = store.GetPath (iter);
	}
	
	public void ShowNode (Node n)
	{
		if (node_to_iter [n] == null){
			OpenTree (n);
			if (node_to_iter [n] == null){
				Console.Error.WriteLine ("Internal error: no node to iter mapping");
				return;
			}
		}
		
		TreeIter iter = (TreeIter) node_to_iter [n];
		TreePath path = store.GetPath (iter);

		//
		// HACK until we upgrade Gtk# to Gtk 2.2
		//
		int depth = path.Depth;
		for (int i = 0; i < depth; i++){
			TreePath p = store.GetPath (iter);

			for (int j = 0; j < depth-i; j++)
				p.Up ();

			tree_view.ExpandRow (p, false);
		}

		IgnoreRowActivated = true;
		tree_view.Selection.SelectIter (iter);
		IgnoreRowActivated = false;
		tree_view.ScrollToCell (path, null, true, 0.5f, 0.5f);
	}
	
	class NodePageVisit : PageVisit {
		Browser browser;
		Node n;
		string url;

		public NodePageVisit (Browser browser, Node n, string url)
		{
			if (n == null)
				throw new Exception ("N is null");
			
			this.browser = browser;
			this.n = n;
			this.url = url;
		}

		public override void Go ()
		{
			Node x;
			
			string res = n.tree.HelpSource.GetText (url, out x);
			((Browser)browser).Render (res, null, url);
		}
	}

	bool IgnoreRowActivated = false;
	
	//
	// This has to handle two kinds of urls: those encoded in the tree
	// file, which are used to quickly lookup information precisely
	// (things like "ecma:0"), and if that fails, it uses the more expensive
	// mechanism that walks trees to find matches
	//
	void RowActivated  (object sender, EventArgs a)
	{
		if (IgnoreRowActivated)
			return;
		
		Gtk.TreeIter iter = new Gtk.TreeIter ();
		Gtk.TreeModel model;

		if (tree_view.Selection.GetSelected (out model, ref iter)){
			Node n = (Node) iter_to_node [iter];

			if (n.tree.HelpSource == null)
				return;

			string url = n.URL;
			
			//
			// Try the tree-based urls first.
			//
			Node match;
			string s = n.tree.HelpSource.GetText (url, out match);
			if (s != null){
				((Browser)browser).Render (s, match, url);
				browser.history.AppendHistory (new NodePageVisit (browser, n, url));
				return;
			}

			//
			// Try the url resolver next
			//
			s = help_tree.RenderUrl (url, out match);
			if (s != null){
				((Browser)browser).Render (s, match, url);
				browser.history.AppendHistory (new Browser.LinkPageVisit (browser, url));
				return;
			}

			((Browser)browser).Render ("<b>Unhandled URL</b> " + n.URL, null, url);
		}
	}
}
