//
// list.cs: A list widget that scales for lots of data.
//
// Author:
//   Miguel de Icaza (miguel@ximian.com)
//
// (C) 2003 Ximian, Inc.
//
using Gdk;
using Gtk;
using GtkSharp;
using System;
using System.Collections;

delegate void ItemSelected (int index);

class BigList : Gtk.DrawingArea {
	Gtk.Adjustment adjustment;
	int top_displayed;
	Pango.Layout layout;
	IListModel provider;
	int line_height;
	int selected_line = -1;
	
	public BigList (IListModel provider)
	{
		this.provider = provider;

		adjustment = new Gtk.Adjustment (0, 0, provider.Rows, 1, 1, 1);
		adjustment.ValueChanged += new EventHandler (ValueChangedHandler);
		
		layout = new Pango.Layout (PangoContext);
		ExposeEvent += new ExposeEventHandler (ExposeHandler);
		ButtonPressEvent += new ButtonPressEventHandler (ButtonHandler);
		Realized += new EventHandler (RealizeHandler);
		Unrealized += new EventHandler (UnrealizeHandler);
		
		AddEvents ((int) EventMask.ButtonPressMask);

		//
		// Hack to compute the size of the font.
		//
		int width;
		layout.GetPixelSize (out width, out line_height);

	}

	void RealizeHandler (object o, EventArgs sender)
	{
	}

	void UnrealizeHandler (object o, EventArgs sender)
	{
	}
	
	void ValueChangedHandler (object obj, EventArgs e)
	{
		top_displayed = (int) Adjustment.Value;
		Refresh ();
	}

	void RedrawLine (int line)
	{
		QueueDrawArea (0, line * line_height, Allocation.width, (line+1) * line_height);
	}

	public void Refresh ()
	{
		QueueDrawArea (0, 0, Allocation.width, Allocation.height);
	}
	
	void ButtonHandler (object o, ButtonPressEventArgs a)
	{
		Gdk.EventButton pos = a.Event;
		int line = (int) (pos.y / line_height);

		if (line > provider.Rows)
			return;

		if (selected_line != -1)
			RedrawLine (selected_line);
		selected_line = line;
		RedrawLine (selected_line);
		
		if (ItemSelected != null)
			ItemSelected (line + top_displayed);
	}

	public event ItemSelected ItemSelected;
	
        void ExposeHandler (object obj, ExposeEventArgs args)
        {
                Gdk.Window win = args.Event.window;
                Gdk.Rectangle area = args.Event.area;

		for (int y = 0, row = top_displayed; y < area.y + area.height; y += line_height, row++){
			if (row >= provider.Rows)
				break;

			string text = provider.GetValue (row);
			Gdk.GC gc;
			Gdk.Rectangle region_area = new Gdk.Rectangle (0, y, area.width, area.height);
			StateType state;
			if (row == selected_line)
				state = StateType.Selected;
			else
				state = StateType.Normal;
			
			win.DrawRectangle (Style.BackgroundGC (state), 1, region_area);
			gc = Style.ForegroundGC (state);
				
			layout.SetMarkup (text);
			win.DrawLayout (gc, 2, y, layout);
		}
		args.RetVal = true;
	}

	public Gtk.Adjustment Adjustment {
		get {
			return adjustment;
		}
	}
}

